from typing import Optional

from databento.common.bento import Bento, FileBento, MemoryBento
from databento.historical.api import API_VERSION
from databento.historical.client import Historical
from databento.historical.error import (
    BentoClientError,
    BentoError,
    BentoHttpError,
    BentoServerError,
)
from databento.version import __version__  # noqa


__all__ = [
    "API_VERSION",
    "Bento",
    "BentoClientError",
    "BentoError",
    "BentoHttpError",
    "BentoServerError",
    "FileBento",
    "Historical",
    "MemoryBento",
]

# Set to either 'DEBUG' or 'INFO', controls console logging
log: Optional[str] = None



from datetime import date
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from databento.common.enums import (
    Compression,
    Dataset,
    Delivery,
    Duration,
    Encoding,
    Packaging,
    Schema,
    SType,
)
from databento.common.parsing import (
    maybe_datetime_to_string,
    maybe_values_list_to_string,
)
from databento.common.validation import validate_enum
from databento.historical.api import API_VERSION
from databento.historical.http import BentoHttpAPI


class BatchHttpAPI(BentoHttpAPI):
    """
    Provides request methods for the batch HTTP API endpoints.
    """

    def __init__(self, key, gateway):
        super().__init__(key=key, gateway=gateway)
        self._base_url = gateway + f"/v{API_VERSION}/batch"

    def submit_job(
        self,
        dataset: Union[Dataset, str],
        schema: Union[Schema, str],
        symbols: Optional[Union[List[str], str]] = None,
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        encoding: Union[Encoding, str] = "dbz",
        compression: Optional[Union[Compression, str]] = None,
        split_duration: Union[Duration, str] = "day",
        split_size: Optional[int] = None,
        packaging: Union[Packaging, str] = "none",
        delivery: Union[Delivery, str] = "download",
        stype_in: Union[SType, str] = "native",
        stype_out: Union[SType, str] = "product_id",
        limit: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Request a new time series data batch job from Databento.

        Makes a `POST /batch.submit_job` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If more than 1 symbol is specified, the data is merged and sorted by time.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime of the request time range (inclusive).
            Assumes UTC as timezone unless passed a tz-aware object.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime of the request time range (exclusive).
            Assumes UTC as timezone unless passed a tz-aware object.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        encoding : Encoding or str {'dbz', 'csv', 'json'}, default 'dbz'
            The data encoding.
        compression : Compression or str {'none', 'zstd'}, optional
            The data compression format (if any).
            If encoding is 'dbz' then specifying a `compression` is invalid (already zstd compressed).
        split_duration : Duration or str {'day', 'week', 'month', 'none'}, default 'day'
            The maximum time duration before batched data is split into multiple files.
            A week starts on Sunday UTC.
        split_size : int, optional
            The maximum size (bytes) of each batched data file before being split.
        packaging : Packaging or str {'none', 'zip', 'tar'}, default 'none'
            The archive type to package all batched data files in.
        delivery : Delivery or str {'download', 's3', 'disk'}, default 'download'
            The delivery mechanism for the processed batched data files.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        stype_out : SType or str, default 'product_id'
            The output symbology type to resolve to.
        limit : int, optional
            The maximum number of records for the request. If ``None`` then no limit.

        Returns
        -------
        Dict[str, Any]
            The job info for submitted batch data request.

        Warnings
        --------
        Calling this method will incur a cost.

        """
        if compression is None:
            compression = Compression.NONE

        validate_enum(schema, Schema, "schema")
        validate_enum(encoding, Encoding, "encoding")
        validate_enum(compression, Compression, "compression")
        validate_enum(split_duration, Duration, "duration")
        validate_enum(packaging, Packaging, "packaging")
        validate_enum(delivery, Delivery, "delivery")
        validate_enum(stype_in, SType, "stype_in")
        validate_enum(stype_out, SType, "stype_out")

        schema = Schema(schema)
        encoding = Encoding(encoding)
        compression = Compression(compression)
        split_duration = Duration(split_duration)
        packaging = Packaging(packaging)
        delivery = Delivery(delivery)
        stype_in = SType(stype_in)
        stype_out = SType(stype_out)

        params: List[Tuple[str, str]] = BentoHttpAPI._timeseries_params(
            dataset=dataset,
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            limit=limit,
            stype_in=stype_in,
            stype_out=stype_out,
        )

        params.append(("encoding", encoding.value))
        params.append(("compression", compression.value))
        params.append(("split_duration", split_duration.value))
        params.append(("packaging", packaging.value))
        params.append(("delivery", delivery.value))
        if split_size is not None:
            params.append(("split_size", str(split_size)))

        return self._post(
            url=self._base_url + ".submit_job",
            params=params,
            basic_auth=True,
        ).json()

    def list_jobs(
        self,
        states: Optional[Union[List[str], str]] = "received,queued,processing,done",
        since: Optional[Union[pd.Timestamp, date, str, int]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Request all batch job details for the user account.

        The job details will be sorted in order of `ts_received`.

        Makes a `GET /batch.list_jobs` HTTP request.

        Parameters
        ----------
        states : List[str] or str, optional {'received', 'queued', 'processing', 'done', 'expired'}  # noqa
            The filter for jobs states as a list of comma separated values.
        since : pd.Timestamp or date or str or int, optional
            The filter for timestamp submitted (will not include jobs prior to this).

        Returns
        -------
        List[Dict[str, Any]]
            The batch job details.

        """
        states = maybe_values_list_to_string(states)
        since = maybe_datetime_to_string(since)

        params: List[Tuple[str, str]] = [
            ("states", states),
            ("since", since),
        ]

        return self._get(
            url=self._base_url + ".list_jobs",
            params=params,
            basic_auth=True,
        ).json()



import warnings
from datetime import date
from typing import List, Optional, Tuple, Union

import pandas as pd
from databento.common.bento import Bento
from databento.common.enums import Dataset, Encoding, Schema, SType
from databento.common.validation import validate_enum
from databento.historical.api import API_VERSION
from databento.historical.http import BentoHttpAPI


class TimeSeriesHttpAPI(BentoHttpAPI):
    """
    Provides request methods for the time series HTTP API endpoints.
    """

    def __init__(self, key, gateway):
        super().__init__(key=key, gateway=gateway)
        self._base_url = gateway + f"/v{API_VERSION}/timeseries"

    def stream(
        self,
        dataset: Union[Dataset, str],
        symbols: Optional[Union[List[str], str]] = None,
        schema: Union[Schema, str] = "trades",
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        stype_in: Union[SType, str] = "native",
        stype_out: Union[SType, str] = "product_id",
        limit: Optional[int] = None,
        path: Optional[str] = None,
    ) -> Bento:
        """
        Request a historical time series data stream from Databento.

        Makes a `GET /timeseries.stream` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If more than 1 symbol is specified, the data is merged and sorted by time.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime (UTC) of the request time range (inclusive).
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime (UTC) of the request time range (exclusive).
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        stype_out : SType or str, default 'product_id'
            The output symbology type to resolve to.
        limit : int, optional
            The maximum number of records to return. If ``None`` then no limit.
        path : str, optional
            The path to stream the data to on disk (will then return a `FileBento`).

        Returns
        -------
        Bento
            If `path` provided then `FileBento`, otherwise `MemoryBento`.

        Notes
        -----
        The Databento Binary Encoding + Zstd Compression (DBZ) will be streamed.

        Warnings
        --------
        Calling this method will incur a cost.

        """
        validate_enum(schema, Schema, "schema")
        validate_enum(stype_in, SType, "stype_in")
        validate_enum(stype_out, SType, "stype_out")

        schema = Schema(schema)
        stype_in = SType(stype_in)
        stype_out = SType(stype_out)

        params: List[Tuple[str, str]] = BentoHttpAPI._timeseries_params(
            dataset=dataset,
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            stype_in=stype_in,
            stype_out=stype_out,
            limit=limit,
        )

        params.append(("encoding", Encoding.DBZ.value))  # Always requests DBZ

        self._pre_check_data_size(
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            limit=limit,
        )

        bento: Bento = self._create_bento(path=path)

        self._stream(
            url=self._base_url + ".stream",
            params=params,
            basic_auth=True,
            bento=bento,
        )

        return bento

    async def stream_async(
        self,
        dataset: Union[Dataset, str],
        symbols: Optional[Union[List[str], str]] = None,
        schema: Union[Schema, str] = "trades",
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        stype_in: Union[SType, str] = "native",
        stype_out: Union[SType, str] = "product_id",
        limit: Optional[int] = None,
        path: Optional[str] = None,
    ) -> Bento:
        """
        Request a historical time series data stream from Databento asynchronously.

        Makes a `GET /timeseries.stream` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If more than 1 symbol is specified, the data is merged and sorted by time.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime (UTC) of the request time range (inclusive).
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime (UTC) of the request time range (exclusive).
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        stype_out : SType or str, default 'product_id'
            The output symbology type to resolve to.
        limit : int, optional
            The maximum number of records to return. If ``None`` then no limit.
        path : str, optional
            The path to stream the data to on disk (will then return a `FileBento`).

        Returns
        -------
        Bento
            If `path` provided then `FileBento`, otherwise `MemoryBento`.

        Notes
        -----
        The Databento Binary Encoding + Zstd Compression (DBZ) will be streamed.

        Warnings
        --------
        Calling this method will incur a cost.

        """
        validate_enum(schema, Schema, "schema")
        validate_enum(stype_in, SType, "stype_in")
        validate_enum(stype_out, SType, "stype_out")

        schema = Schema(schema)
        stype_in = SType(stype_in)
        stype_out = SType(stype_out)

        params: List[Tuple[str, str]] = BentoHttpAPI._timeseries_params(
            dataset=dataset,
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            stype_in=stype_in,
            stype_out=stype_out,
            limit=limit,
        )

        params.append(("encoding", Encoding.DBZ.value))  # Always requests DBZ

        self._pre_check_data_size(
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            limit=limit,
        )

        bento: Bento = self._create_bento(path=path)

        await self._stream_async(
            url=self._base_url + ".stream",
            params=params,
            basic_auth=True,
            bento=bento,
        )

        return bento

    def _pre_check_data_size(
        self,
        symbols: Optional[Union[List[str], str]],
        schema: Schema,
        start: Optional[Union[pd.Timestamp, date, str, int]],
        end: Optional[Union[pd.Timestamp, date, str, int]],
        limit: Optional[int],
    ):
        if limit and limit < 10**7:
            return

        # Use heuristics to check ballpark data size
        if (
            _is_large_data_size_schema(schema)
            or _is_greater_than_one_day(start, end)
            or _is_large_number_of_symbols(symbols)
        ):
            warnings.warn(
                "\nThe size of the current streaming request is estimated "
                "to be 5 GB or greater. We recommend smaller "
                "individual streaming request sizes, or alternatively "
                "submit a batch data request."
                "\nYou can check the uncompressed binary size of a request "
                "through the metadata API (from the client library, or over "
                "HTTP).\nThis warning can be suppressed "
                "https://docs.python.org/3/library/warnings.html",
            )


def _is_large_number_of_symbols(symbols: Optional[Union[List[str], str]]):
    if not symbols:
        return True  # All symbols

    if isinstance(symbols, str):
        symbols = symbols.split(",")

    if len(symbols) >= 500:
        return True

    return False


def _is_large_data_size_schema(schema: Schema):
    return schema in (Schema.MBO, Schema.MBP_10)


def _is_greater_than_one_day(start, end):
    if start is None or end is None:
        return True

    if pd.to_datetime(end) - pd.to_datetime(start) > pd.Timedelta(days=1):
        return True

    return False



from datetime import date
from enum import Enum
from typing import Iterable, List, Optional, Union

import pandas as pd
from databento.common.enums import Compression, Encoding, Flags, Schema, SType


def enum_or_str_lowercase(
    value: Union[Enum, str],
    param: str,
) -> str:
    """
    Return the given value parsed to a lowercase string if possible.

    Parameters
    ----------
    value : Enum or str
        The value to parse.
    param : str
        The name of the parameter being validated (for any error message).

    Returns
    -------
    str

    Raises
    ------
    TypeError
        If value is not of type NoneType, Enum or str.

    """
    if isinstance(value, Enum):
        return value.value.lower()
    elif isinstance(value, str):
        if not value.isspace():
            return value.lower()

    raise TypeError(f"invalid `{param}` type, was {type(value)}.")


def maybe_enum_or_str_lowercase(
    value: Optional[Union[Enum, str]],
    param: str,
) -> Optional[str]:
    """
    Return the given value parsed to a lowercase string if possible.

    Parameters
    ----------
    value : Enum or str, optional
        The value to parse.
    param : str
        The name of the parameter being validated (for any error message).

    Returns
    -------
    str or ``None``

    Raises
    ------
    TypeError
        If value is not of type NoneType, Enum or str.

    """
    if value is None:
        return value
    return enum_or_str_lowercase(value, param)


def values_list_to_string(
    values: Union[Iterable[str], str],
) -> str:
    """
    Concatenate a values string or iterable of string values.

    Parameters
    ----------
    values : iterable of str or str
        The values to concatenate.

    Returns
    -------
    str

    """
    if isinstance(values, str):
        return values.strip().rstrip(",").lower()
    elif isinstance(values, Iterable):
        return ",".join(values).strip().lower()
    else:
        raise TypeError(f"invalid values type, was {type(values)}")


def maybe_values_list_to_string(
    values: Optional[Union[Iterable[str], str]],
) -> Optional[str]:
    """
    Concatenate a values string or iterable of string values (if not None).

    Parameters
    ----------
    values : iterable of str or str, optional
        The values to concatenate.

    Returns
    -------
    str or ``None``

    """
    if values is None:
        return None

    return values_list_to_string(values)


def maybe_symbols_list_to_string(
    symbols: Optional[Union[Iterable[str], str]],
) -> Optional[str]:
    """
    Concatenate a symbols string or iterable of symbol strings (if not None).

    Parameters
    ----------
    symbols : iterable of str or str, optional
        The symbols to concatenate.

    Returns
    -------
    str or ``None``

    """
    if symbols is None:
        return None  # All symbols

    if isinstance(symbols, str):
        return symbols.strip().rstrip(",").upper()
    elif isinstance(symbols, Iterable):
        return ",".join(symbols).strip().upper()
    else:
        raise TypeError(f"invalid symbols type, was {type(symbols)}")


def maybe_date_to_string(
    value: Optional[Union[date, str]],
) -> Optional[str]:
    """
    Return a valid date string from the given value (if not None).

    Parameters
    ----------
    value : date or str, optional
        The value to parse.

    Returns
    -------
    str or ``None``

    """
    if value is None:
        return None

    return str(pd.to_datetime(value).date())


def maybe_datetime_to_string(
    value: Optional[Union[pd.Timestamp, date, str]],
) -> Optional[str]:
    """
    Return a valid datetime string from the given value (if not None).

    Parameters
    ----------
    value : pd.Timestamp or date or str, optional
        The value to parse.

    Returns
    -------
    str or ``None``

    """
    if value is None:
        return None

    return str(pd.to_datetime(value)).replace(" ", "T")


def parse_flags(value: int, apply_bitmask: bool = False) -> List[str]:
    """
    Return an array of flag strings from the given flags value.

    This convenience function parses the values from the flags field.
    The flags value is a combination of event packet end and matching engine
    status. Using a technique involving bit shifts, there can be more than one
    flag represented by a single integer.

    Possible values include:
     - F_LAST: Last msg in packet (flags < 0).
     - F_HALT: Exchange-independent HALT signal.
     - F_RESET: Drop book, reset symbol for this exchange.
     - F_DUPID: This OrderID has valid fresh duplicate (Iceberg, etc).
     - F_MBP: This is SIP/MBP ADD message, single per price level.
     - F_RESERVED2: Reserved for future use (no current meaning).
     - F_RESERVED1: Reserved for future use (no current meaning).
     - F_RESERVED0: Reserved for future use (no current meaning).

    Parameters
    ----------
    value : int
        The 'flags' field value.
    apply_bitmask : bool, default False
        If the AND 0xff bitmask should be applied to remove the values sign.
        This would only be required when parsing raw binary data.

    Returns
    -------
    list[str]

    """
    if apply_bitmask:
        value = value & 0xFF
    return [f.name for f in Flags if value & f.value != 0]


# TODO(cs): We should probably change the enum values to ints so as to avoid
#  all of these conversion functions. If this implementation remains then will
#  add exhaustive unit tests (27/6/22)


def schema_to_int(schema: Schema) -> int:
    assert isinstance(schema, Schema)

    if schema == Schema.MBO:
        return 0
    elif schema == Schema.MBP_1:
        return 1
    elif schema == Schema.MBP_10:
        return 2
    elif schema == Schema.TBBO:
        return 3
    elif schema == Schema.TRADES:
        return 4
    elif schema == Schema.OHLCV_1S:
        return 5
    elif schema == Schema.OHLCV_1M:
        return 6
    elif schema == Schema.OHLCV_1H:
        return 7
    elif schema == Schema.OHLCV_1D:
        return 8
    elif schema == Schema.DEFINITION:
        return 9
    elif schema == Schema.STATISTICS:
        return 10
    elif schema == Schema.STATUS:
        return 11
    else:
        raise NotImplementedError(
            f"The enum value '{schema.value}' "
            f"has not been implemented for conversion",
        )


def int_to_schema(value: int) -> Schema:
    if value == 0:
        return Schema.MBO
    elif value == 1:
        return Schema.MBP_1
    elif value == 2:
        return Schema.MBP_10
    elif value == 3:
        return Schema.TBBO
    elif value == 4:
        return Schema.TRADES
    elif value == 5:
        return Schema.OHLCV_1S
    elif value == 6:
        return Schema.OHLCV_1M
    elif value == 7:
        return Schema.OHLCV_1H
    elif value == 8:
        return Schema.OHLCV_1D
    elif value == 9:
        return Schema.DEFINITION
    elif value == 10:
        return Schema.STATISTICS
    elif value == 11:
        return Schema.STATUS
    else:
        raise NotImplementedError(
            f"The int value '{value}' " f"cannot be represented with the enum",
        )


def stype_to_int(stype: SType) -> int:
    assert isinstance(stype, SType)

    if stype == SType.PRODUCT_ID:
        return 0
    elif stype == SType.NATIVE:
        return 1
    elif stype == SType.SMART:
        return 2
    else:
        raise NotImplementedError(
            f"The enum value '{stype.value}' "
            f"has not been implemented for conversion",
        )


def int_to_stype(value: int) -> SType:
    if value == 0:
        return SType.PRODUCT_ID
    elif value == 1:
        return SType.NATIVE
    elif value == 2:
        return SType.SMART
    else:
        raise NotImplementedError(
            f"The int value '{value}' " f"cannot be represented with the enum",
        )


def encoding_to_int(encoding: Encoding) -> int:
    assert isinstance(encoding, Encoding)

    if encoding == Encoding.DBZ:
        return 0
    elif encoding == Encoding.CSV:
        return 1
    elif encoding == Encoding.JSON:
        return 2
    else:
        raise NotImplementedError(
            f"The enum value '{encoding.value}' "
            f"has not been implemented for conversion",
        )


def int_to_encoding(value: int) -> Encoding:
    if value == 0:
        return Encoding.DBZ
    elif value == 1:
        return Encoding.CSV
    elif value == 2:
        return Encoding.JSON
    else:
        raise NotImplementedError(
            f"The int value '{value}' " f"cannot be represented with the enum",
        )


def compression_to_int(compression: Compression) -> int:
    assert isinstance(compression, Compression)

    if compression == Compression.NONE:
        return 0
    elif compression == Compression.ZSTD:
        return 1
    else:
        raise NotImplementedError(
            f"The enum value '{compression.value}' "
            f"has not been implemented for conversion",
        )


def int_to_compression(value: int) -> Compression:
    if value == 0:
        return Compression.NONE
    elif value == 1:
        return Compression.ZSTD
    else:
        raise NotImplementedError(
            f"The int value '{value}' " f"cannot be represented with the enum",
        )



from typing import Dict, List, Tuple, Union

import numpy as np
from databento.common.enums import Schema


################################################################################
# DBZ struct schema
################################################################################


def get_deriv_ba_types(level: int) -> List[Tuple[str, Union[type, str]]]:
    return [
        (f"bid_px_{level:02d}", np.int64),
        (f"ask_px_{level:02d}", np.int64),
        (f"bid_sz_{level:02d}", np.uint32),
        (f"ask_sz_{level:02d}", np.uint32),
        (f"bid_oq_{level:02d}", np.uint32),
        (f"ask_oq_{level:02d}", np.uint32),
    ]


DERIV_SCHEMAS = (
    Schema.MBP_1,
    Schema.MBP_10,
    Schema.TBBO,
    Schema.TRADES,
)


OHLCV_SCHEMAS = (
    Schema.OHLCV_1S,
    Schema.OHLCV_1M,
    Schema.OHLCV_1H,
    Schema.OHLCV_1D,
)


DBZ_COMMON_HEADER: List[Tuple[str, Union[type, str]]] = [
    ("nwords", np.uint8),
    ("type", np.uint8),
    ("publisher_id", np.uint16),
    ("product_id", np.uint32),
    ("ts_event", np.uint64),
]


DBZ_MBP_MSG: List[Tuple[str, Union[type, str]]] = [
    ("price", np.int64),
    ("size", np.uint32),
    ("action", "S1"),  # 1 byte chararray
    ("side", "S1"),  # 1 byte chararray
    ("flags", np.int8),
    ("depth", np.uint8),
    ("ts_recv", np.uint64),
    ("ts_in_delta", np.int32),
    ("sequence", np.uint32),
]


DBZ_OHLCV_MSG: List[Tuple[str, Union[type, str]]] = [
    ("open", np.int64),
    ("high", np.int64),
    ("low", np.int64),
    ("close", np.int64),
    ("volume", np.int64),
]


DBZ_STRUCT_MAP: Dict[Schema, List[Tuple[str, Union[type, str]]]] = {
    Schema.MBO: DBZ_COMMON_HEADER
    + [
        ("order_id", np.uint64),
        ("price", np.int64),
        ("size", np.uint32),
        ("flags", np.int8),
        ("channel_id", np.uint8),
        ("action", "S1"),  # 1 byte chararray
        ("side", "S1"),  # 1 byte chararray
        ("ts_recv", np.uint64),
        ("ts_in_delta", np.int32),
        ("sequence", np.uint32),
    ],
    Schema.MBP_1: DBZ_COMMON_HEADER + DBZ_MBP_MSG + get_deriv_ba_types(0),  # 1
    Schema.MBP_10: DBZ_COMMON_HEADER
    + DBZ_MBP_MSG
    + get_deriv_ba_types(0)  # 1
    + get_deriv_ba_types(1)  # 2
    + get_deriv_ba_types(2)  # 3
    + get_deriv_ba_types(3)  # 4
    + get_deriv_ba_types(4)  # 5
    + get_deriv_ba_types(5)  # 6
    + get_deriv_ba_types(6)  # 7
    + get_deriv_ba_types(7)  # 8
    + get_deriv_ba_types(8)  # 9
    + get_deriv_ba_types(9),  # 10
    Schema.TBBO: DBZ_COMMON_HEADER + DBZ_MBP_MSG + get_deriv_ba_types(0),
    Schema.TRADES: DBZ_COMMON_HEADER + DBZ_MBP_MSG,
    Schema.OHLCV_1S: DBZ_COMMON_HEADER + DBZ_OHLCV_MSG,
    Schema.OHLCV_1M: DBZ_COMMON_HEADER + DBZ_OHLCV_MSG,
    Schema.OHLCV_1H: DBZ_COMMON_HEADER + DBZ_OHLCV_MSG,
    Schema.OHLCV_1D: DBZ_COMMON_HEADER + DBZ_OHLCV_MSG,
    Schema.STATUS: DBZ_COMMON_HEADER
    + [
        ("ts_recv", np.uint64),
        ("group", "S1"),  # 1 byte chararray
        ("trading_status", np.uint8),
        ("halt_reason", np.uint8),
        ("trading_event", np.uint8),
    ],
    Schema.DEFINITION: DBZ_COMMON_HEADER
    + [
        ("ts_recv", np.uint64),
        ("min_price_increment", np.int64),
        ("display_factor", np.int64),
        ("expiration", np.uint64),
        ("activation", np.uint64),
        ("high_limit_price", np.int64),
        ("low_limit_price", np.int64),
        ("max_price_variation", np.int64),
        ("trading_reference_price", np.int64),
        ("unit_of_measure_qty", np.int64),
        ("min_price_increment_amount", np.int64),
        ("price_ratio", np.int64),
        ("inst_attrib_value", np.int32),
        ("underlying_id", np.uint32),
        ("cleared_volume", np.int32),
        ("market_depth_implied", np.int32),
        ("market_depth", np.int32),
        ("market_segment_id", np.uint32),
        ("max_trade_vol", np.uint32),
        ("min_lot_size", np.int32),
        ("min_lot_size_block", np.int32),
        ("min_lot_size_round_lot", np.int32),
        ("min_trade_vol", np.uint32),
        ("open_interest_qty", np.int32),
        ("contract_multiplier", np.int32),
        ("decay_quantity", np.int32),
        ("original_contract_size", np.int32),
        ("related_security_id", np.uint32),
        ("trading_reference_date", np.uint16),
        ("appl_id", np.int16),
        ("maturity_month_year", np.uint16),
        ("decay_start_date", np.uint16),
        ("chan", np.uint16),
        ("currency", "S1"),  # 1 byte chararray
        ("settl_currency", "S1"),  # 1 byte chararray
        ("secsubtype", "S1"),  # 1 byte chararray
        ("symbol", "S1"),  # 1 byte chararray
        ("group", "S1"),  # 1 byte chararray
        ("exchange", "S1"),  # 1 byte chararray
        ("asset", "S1"),  # 1 byte chararray
        ("cfi", "S1"),  # 1 byte chararray
        ("security_type", "S1"),  # 1 byte chararray
        ("unit_of_measure", "S1"),  # 1 byte chararray
        ("underlying", "S1"),  # 1 byte chararray
        ("related", "S1"),  # 1 byte chararray
        ("match_algorithm", "S1"),  # 1 byte chararray
        ("md_security_trading_status", np.uint8),
        ("main_fraction", np.uint8),
        ("price_display_format", np.uint8),
        ("settl_price_type", np.uint8),
        ("sub_fraction", np.uint8),
        ("underlying_product", np.uint8),
        ("security_update_action", "S1"),  # 1 byte chararray
        ("maturity_month_month", np.uint8),
        ("maturity_month_day", np.uint8),
        ("maturity_month_week", np.uint8),
        ("user_defined_instrument", "S1"),  # 1 byte chararray
        ("contract_multiplier_unit", np.int8),
        ("flow_schedule_type", np.int8),
        ("tick_rule", np.uint8),
        ("dummy", "S1"),  # 1 byte chararray
    ],
}


################################################################################
# DBZ fields
################################################################################


def get_deriv_ba_fields(level: int) -> List[str]:
    return [
        f"bid_px_{level:02d}",
        f"ask_px_{level:02d}",
        f"bid_sz_{level:02d}",
        f"ask_sz_{level:02d}",
        f"bid_oq_{level:02d}",
        f"ask_oq_{level:02d}",
    ]


DBZ_DERIV_HEADER_FIELDS = [
    "ts_event",
    "ts_in_delta",
    "publisher_id",
    "product_id",
    "action",
    "side",
    "flags",
    "price",
    "size",
    "sequence",
]

DBZ_COLUMNS = {
    Schema.MBO: [
        "ts_event",
        "ts_in_delta",
        "publisher_id",
        "product_id",
        "order_id",
        "action",
        "side",
        "flags",
        "price",
        "size",
        "sequence",
    ],
    Schema.MBP_1: DBZ_DERIV_HEADER_FIELDS + get_deriv_ba_fields(0),
    Schema.MBP_10: DBZ_DERIV_HEADER_FIELDS
    + get_deriv_ba_fields(0)
    + get_deriv_ba_fields(1)
    + get_deriv_ba_fields(2)
    + get_deriv_ba_fields(3)
    + get_deriv_ba_fields(4)
    + get_deriv_ba_fields(5)
    + get_deriv_ba_fields(6)
    + get_deriv_ba_fields(7)
    + get_deriv_ba_fields(8)
    + get_deriv_ba_fields(9),
    Schema.TBBO: DBZ_DERIV_HEADER_FIELDS + get_deriv_ba_fields(0),
    Schema.TRADES: DBZ_DERIV_HEADER_FIELDS,
}



import logging
import os
import sys

import databento


DATABENTO_LOG = os.environ.get("DATABENTO_LOG", "").upper()

logger = logging.getLogger("databento")

_DEBUG = "DEBUG"
_INFO = "INFO"
_ERROR = "ERROR"


def _console_log_level():
    if databento.log:
        databento.log = databento.log.upper()

    if databento.log in [_DEBUG, _INFO, _ERROR]:
        return databento.log
    elif DATABENTO_LOG in [_DEBUG, _INFO, _ERROR]:
        return DATABENTO_LOG
    else:
        return None


def log_debug(msg: str):
    """Log the given message with DEBUG level."""
    log_level = _console_log_level()
    if log_level == [_DEBUG, _INFO, _ERROR]:
        print(f"DEBUG: {msg}", file=sys.stderr)
    logger.debug(msg)


def log_info(msg: str):
    """Log the given message with INFO level."""
    log_level = _console_log_level()
    if log_level in [_INFO, _ERROR]:
        print(f"INFO: {msg}", file=sys.stderr)
    logger.info(msg)


def log_error(msg: str):
    """Log the given message with ERROR level."""
    log_level = _console_log_level()
    if log_level in [_ERROR]:
        print(f"ERROR: {msg}", file=sys.stderr)
    logger.error(msg)



from enum import Enum, unique


@unique
class HistoricalGateway(Enum):
    """Represents a historical data center gateway location."""

    NEAREST = "nearest"
    BO1 = "bo1"


@unique
class LiveGateway(Enum):
    """Represents a live data center gateway location."""

    ORIGIN = "origin"
    NEAREST = "nearest"
    NY4 = "ny4"
    DC3 = "dc3"


@unique
class FeedMode(Enum):
    """Represents a data feed mode."""

    HISTORICAL = "historical"
    HISTORICAL_STREAMING = "historical-streaming"
    LIVE = "live"


@unique
class Dataset(Enum):
    """Represents a dataset code (string identifier)."""

    GLBX_MDP3 = "GLBX.MDP3"
    XNAS_ITCH = "XNAS.ITCH"


@unique
class Schema(Enum):
    """Represents a data record schema."""

    MBO = "mbo"
    MBP_1 = "mbp-1"
    MBP_10 = "mbp-10"
    TBBO = "tbbo"
    TRADES = "trades"
    OHLCV_1S = "ohlcv-1s"
    OHLCV_1M = "ohlcv-1m"
    OHLCV_1H = "ohlcv-1h"
    OHLCV_1D = "ohlcv-1d"
    DEFINITION = "definition"
    STATISTICS = "statistics"
    STATUS = "status"


@unique
class Encoding(Enum):
    """Represents a data output encoding."""

    DBZ = "dbz"
    CSV = "csv"
    JSON = "json"


@unique
class Compression(Enum):
    """Represents a data compression format (if any)."""

    NONE = "none"
    ZSTD = "zstd"


@unique
class Duration(Enum):
    """Represents the duration interval for each batch data file."""

    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    NONE = "none"


@unique
class Packaging(Enum):
    """Represents the packaging method for batched data files."""

    NONE = "none"
    ZIP = "zip"
    TAR = "tar"


@unique
class Delivery(Enum):
    """Represents the delivery mechanism for batch data."""

    DOWNLOAD = "download"
    S3 = "s3"
    DISK = "disk"


@unique
class SType(Enum):
    """Represents a symbology type."""

    PRODUCT_ID = "product_id"
    NATIVE = "native"
    SMART = "smart"


@unique
class RollRule(Enum):
    """Represents a smart symbology roll rule."""

    VOLUME = 0
    OPEN_INTEREST = 1
    CALENDAR = 2


@unique
class SymbologyResolution(Enum):
    """
    Status code of symbology resolution.

    - OK: All symbol mappings resolved.
    - PARTIAL: One or more symbols did not resolve on at least one date.
    - NOT_FOUND: One or more symbols where not found on any date in range.
    """

    OK = 0
    PARTIAL = 1
    NOT_FOUND = 2


@unique
class Flags(Enum):
    """Represents record flags."""

    F_LAST = 1 << 7  # 128  Last msg in packet (flags < 0)
    F_HALT = 1 << 6  # 64  Exchange-independent HALT signal
    F_RESET = 1 << 5  # 32  Drop book, reset symbol for this exchange
    F_DUPID = 1 << 4  # 16  This OrderID has valid fresh duplicate (Iceberg, etc)
    F_MBP = 1 << 3  # 8  This is SIP/MBP ADD message, single per price level
    F_RESERVED2 = 1 << 2  # 4 Reserved for future use
    F_RESERVED1 = 1 << 1  # 2 Reserved for future use
    F_RESERVED0 = 1  # Reserved for future use



import os
from typing import Optional, Union

from databento.common.enums import HistoricalGateway
from databento.common.logging import log_info
from databento.common.parsing import enum_or_str_lowercase
from databento.historical.api.batch import BatchHttpAPI
from databento.historical.api.metadata import MetadataHttpAPI
from databento.historical.api.symbology import SymbologyHttpAPI
from databento.historical.api.timeseries import TimeSeriesHttpAPI


class Historical:
    """
    Provides a client connection class for requesting historical data and
    metadata from the Databento API servers.

    Parameters
    ----------
    key : str, optional
        The user API key for authentication.
        If ``None`` then the `DATABENTO_API_KEY` environment variable is used.
    gateway : HistoricalGateway or str, default HistoricalGateway.NEAREST
        The API server gateway.
        If ``None`` then the default gateway is used.

    Examples
    --------
    > import databento as db
    > client = db.Historical('YOUR_API_KEY')
    """

    def __init__(
        self,
        key: Optional[str] = None,
        gateway: Union[HistoricalGateway, str] = HistoricalGateway.NEAREST,
    ):
        if key is None:
            key = os.environ.get("DATABENTO_API_KEY")
        if key is None or not isinstance(key, str) or key.isspace():
            raise ValueError(f"invalid API key, was {key}")

        # Configure data access gateway
        gateway = enum_or_str_lowercase(gateway, "gateway")
        if gateway in ("nearest", "bo1"):
            gateway = "https://hist.databento.com"

        self._key = key
        self._gateway = gateway

        self.batch = BatchHttpAPI(key=key, gateway=gateway)
        self.metadata = MetadataHttpAPI(key=key, gateway=gateway)
        self.symbology = SymbologyHttpAPI(key=key, gateway=gateway)
        self.timeseries = TimeSeriesHttpAPI(key=key, gateway=gateway)

        # Not logging security sensitive `key`
        log_info(f"Initialized {type(self).__name__}(gateway={self._gateway})")

    @property
    def key(self) -> str:
        """
        Return the user API key for the client.

        Returns
        -------
        str

        """
        return self._key

    @property
    def gateway(self) -> str:
        """
        Return the API server gateway for the client.

        Returns
        -------
        str

        """
        return self._gateway



from typing import Any, Dict

from databento.common.parsing import int_to_compression, int_to_schema, int_to_stype
from dbz_lib import decode_metadata


class MetadataDecoder:
    """
    Provides a decoder for DBZ metadata headers.
    """

    @staticmethod
    def decode_to_json(raw_metadata: bytes) -> Dict[str, Any]:
        """
        Decode the given metadata into a JSON object (as a Python dict).

        Parameters
        ----------
        raw_metadata : bytes
            The metadata to decode.

        Returns
        -------
        Dict[str, Any]

        """

        def enum_value(fn):
            return lambda x: fn(x).value

        metadata = decode_metadata(raw_metadata)
        conversion_mapping = {
            "compression": enum_value(int_to_compression),
            "limit": lambda lim: None if lim == 0 else lim,
            "mappings": lambda m: {i["native"]: i["intervals"] for i in m},
            "schema": enum_value(int_to_schema),
            "stype_in": enum_value(int_to_stype),
            "stype_out": enum_value(int_to_stype),
        }
        for key, conv_fn in conversion_mapping.items():
            metadata[key] = conv_fn(metadata[key])
        return metadata



import sys
from datetime import date
from json.decoder import JSONDecodeError
from typing import Any, BinaryIO, List, Optional, Tuple, Union

import aiohttp
import pandas as pd
import requests
from aiohttp import ClientResponse
from databento.common.bento import Bento, FileBento, MemoryBento
from databento.common.enums import Dataset, Schema, SType
from databento.common.logging import log_info
from databento.common.parsing import (
    enum_or_str_lowercase,
    maybe_datetime_to_string,
    maybe_symbols_list_to_string,
)
from databento.historical.error import BentoClientError, BentoServerError
from databento.version import __version__
from requests import Response
from requests.auth import HTTPBasicAuth


_NO_DATA_FOUND = b"No data found for query."
_32KB = 1024 * 32  # 32_768


class BentoHttpAPI:
    """The base class for all Databento HTTP API endpoints."""

    TIMEOUT = 100

    def __init__(self, key: str, gateway: str):
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
        user_agent = f"Databento/{__version__} Python/{python_version}"

        self._key = key
        self._gateway = gateway
        self._headers = {"accept": "application/json", "user-agent": user_agent}

    @staticmethod
    def _timeseries_params(
        *,
        dataset: Union[Dataset, str],
        symbols: Optional[Union[List[str], str]] = None,
        schema: Schema,
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        limit: Optional[int] = None,
        stype_in: SType,
        stype_out: SType = SType.PRODUCT_ID,
    ) -> List[Tuple[str, str]]:
        # Parse inputs
        dataset = enum_or_str_lowercase(dataset, "dataset")
        symbols = maybe_symbols_list_to_string(symbols)
        start = maybe_datetime_to_string(start)
        end = maybe_datetime_to_string(end)

        params: List[Tuple[str, Any]] = [
            ("dataset", dataset),
            ("symbols", symbols),
            ("schema", schema.value),
            ("start", start),
            ("end", end),
            ("stype_in", stype_in.value),
            ("stype_out", stype_out.value),
        ]

        if limit is not None:
            params.append(("limit", str(limit)))

        return params

    @staticmethod
    def _create_bento(path: str) -> Union[MemoryBento, FileBento]:
        if path is None:
            return MemoryBento()
        else:
            return FileBento(path=path)

    def _check_api_key(self):
        if self._key == "YOUR_API_KEY":
            raise ValueError(
                "The API key is currently set to 'YOUR_API_KEY'. "
                "Please replace this value with a valid API key. "
                "You can find these through your Databento user portal.",
            )

    def _get(
        self,
        url: str,
        params: Optional[List[Tuple[str, str]]] = None,
        basic_auth: bool = False,
    ) -> Response:
        self._check_api_key()

        with requests.get(
            url=url,
            params=params,
            headers=self._headers,
            auth=HTTPBasicAuth(username=self._key, password=None)
            if basic_auth
            else None,
            timeout=(self.TIMEOUT, self.TIMEOUT),
        ) as response:
            check_http_error(response)
            return response

    async def _get_async(
        self,
        url: str,
        params: Optional[List[Tuple[str, str]]] = None,
        basic_auth: bool = False,
    ) -> ClientResponse:
        self._check_api_key()
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=url,
                params=params,
                headers=self._headers,
                auth=aiohttp.BasicAuth(login=self._key, password="", encoding="utf-8")
                if basic_auth
                else None,
                timeout=self.TIMEOUT,
            ) as response:
                await check_http_error_async(response)
                return response

    def _post(
        self,
        url: str,
        params: Optional[List[Tuple[str, str]]] = None,
        basic_auth: bool = False,
    ) -> Response:
        self._check_api_key()

        with requests.post(
            url=url,
            params=params,
            headers=self._headers,
            auth=HTTPBasicAuth(username=self._key, password=None)
            if basic_auth
            else None,
            timeout=(self.TIMEOUT, self.TIMEOUT),
        ) as response:
            check_http_error(response)
            return response

    def _stream(
        self,
        url: str,
        params: List[Tuple[str, str]],
        basic_auth: bool,
        bento: Bento,
    ) -> None:
        self._check_api_key()

        with requests.get(
            url=url,
            params=params,
            headers=self._headers,
            auth=HTTPBasicAuth(username=self._key, password=None)
            if basic_auth
            else None,
            timeout=(self.TIMEOUT, self.TIMEOUT),
            stream=True,
        ) as response:
            check_http_error(response)

            # Setup bento I/O writer
            writer: BinaryIO = bento.writer()

            for chunk in response.iter_content(chunk_size=_32KB):
                if chunk == _NO_DATA_FOUND:
                    log_info("No data found for query.")
                    return
                writer.write(chunk)

            if isinstance(bento, FileBento):
                writer.close()

            metadata = bento.source_metadata()
            bento.set_metadata(metadata)

    async def _stream_async(
        self,
        url: str,
        params: List[Tuple[str, Optional[str]]],
        basic_auth: bool,
        bento: Bento,
    ) -> None:
        self._check_api_key()

        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=url,
                params=[x for x in params if x[1] is not None],
                headers=self._headers,
                auth=aiohttp.BasicAuth(login=self._key, password="", encoding="utf-8")
                if basic_auth
                else None,
                timeout=self.TIMEOUT,
            ) as response:
                await check_http_error_async(response)

                # Setup bento I/O writer
                writer: BinaryIO = bento.writer()

                async for chunk in response.content.iter_chunks():
                    data: bytes = chunk[0]
                    if data == _NO_DATA_FOUND:
                        log_info("No data found for query.")
                        return
                    writer.write(data)

                if isinstance(bento, FileBento):
                    writer.close()

                metadata = bento.source_metadata()
                bento.set_metadata(metadata)


def is_400_series_error(status: int) -> bool:
    return status // 100 == 4


def is_500_series_error(status: int) -> bool:
    return status // 100 == 5


def check_http_error(response: Response) -> None:
    if is_500_series_error(response.status_code):
        try:
            json_body = response.json()
            message = json_body.get("detail")
        except JSONDecodeError:
            json_body = None
            message = None
        raise BentoServerError(
            http_status=response.status_code,
            http_body=response.content,
            json_body=json_body,
            message=message,
            headers=response.headers,
        )
    elif is_400_series_error(response.status_code):
        try:
            json_body = response.json()
            message = json_body.get("detail")
        except JSONDecodeError:
            json_body = None
            message = None
        raise BentoClientError(
            http_status=response.status_code,
            http_body=response.content,
            json_body=json_body,
            message=message,
            headers=response.headers,
        )


async def check_http_error_async(response: ClientResponse) -> None:
    if is_500_series_error(response.status):
        json_body = await response.json()
        raise BentoServerError(
            http_status=response.status,
            http_body=response.content,
            json_body=json_body,
            message=json_body["detail"],
            headers=response.headers,
        )
    elif is_400_series_error(response.status):
        json_body = await response.json()
        raise BentoClientError(
            http_status=response.status,
            http_body=response.content,
            json_body=json_body,
            message=json_body["detail"],
            headers=response.headers,
        )



from datetime import date
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from databento.common.enums import Dataset, Encoding, FeedMode, Schema, SType
from databento.common.parsing import (
    enum_or_str_lowercase,
    maybe_date_to_string,
    maybe_datetime_to_string,
    maybe_enum_or_str_lowercase,
    maybe_symbols_list_to_string,
)
from databento.common.validation import validate_enum, validate_maybe_enum
from databento.historical.api import API_VERSION
from databento.historical.http import BentoHttpAPI
from requests import Response


class MetadataHttpAPI(BentoHttpAPI):
    """
    Provides request methods for the metadata HTTP API endpoints.
    """

    def __init__(self, key, gateway):
        super().__init__(key=key, gateway=gateway)
        self._base_url = gateway + f"/v{API_VERSION}/metadata"

    def list_publishers(self) -> Dict[str, int]:
        """
        Request all publishers from Databento.

        Makes a `GET /metadata.list_publishers` HTTP request.

        Use this method to list the mappings of publisher names to publisher IDs.

        Returns
        -------
        Dict[str, int]

        """
        response: Response = self._get(
            url=self._base_url + ".list_publishers",
            basic_auth=True,
        )
        return response.json()

    def list_datasets(
        self,
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
    ) -> List[str]:
        """
        Request all available datasets from Databento.

        Makes a `GET /metadata.list_datasets` HTTP request.

        Use this method to list the _names_ of all available datasets, so you
        can use other methods which take the `dataset` parameter.

        Parameters
        ----------
        start_date : date or str, optional
            The start date (UTC) for the request range.
        end_date : date or str, optional
            The end date (UTC) for the request range.

        Returns
        -------
        List[str]

        """
        start_date = maybe_date_to_string(start_date)
        end_date = maybe_date_to_string(end_date)

        params: List[Tuple[str, str]] = [
            ("start_date", start_date),
            ("end_date", end_date),
        ]

        response: Response = self._get(
            url=self._base_url + ".list_datasets",
            params=params,
            basic_auth=True,
        )
        return response.json()

    def list_schemas(
        self,
        dataset: Union[Dataset, str],
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
    ) -> List[str]:
        """
        Request all available data schemas from Databento.

        Makes a `GET /metadata.list_schemas` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        start_date : date or str, optional
            The start date (UTC) for the request range.
        end_date : date or str, optional
            The end date (UTC) for the request range.

        Returns
        -------
        List[str]

        """
        dataset = enum_or_str_lowercase(dataset, "dataset")
        start_date = maybe_date_to_string(start_date)
        end_date = maybe_date_to_string(end_date)

        params: List[Tuple[str, str]] = [
            ("dataset", dataset),
            ("start_date", start_date),
            ("end_date", end_date),
        ]

        response: Response = self._get(
            url=self._base_url + ".list_schemas",
            params=params,
            basic_auth=True,
        )
        return response.json()

    def list_fields(
        self,
        dataset: Union[Dataset, str],
        schema: Optional[Union[Schema, str]] = None,
        encoding: Optional[Union[Encoding, str]] = None,
    ) -> Dict[str, Dict]:
        """
        Request all fields for a dataset, schema and encoding from Databento.

        Makes a `GET /metadata.list_fields` HTTP request.

        The `schema` and `encoding` parameters act as optional filters. All
        metadata for that parameter is returned if they are not specified.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, optional  # noqa
            The data record schema for the request.
        encoding : Encoding or str {'dbz', 'csv', 'json'}, optional
            The data encoding.

        Returns
        -------
        Dict[str, Any]
            A mapping of dataset to encoding to schema to field to data type.

        """
        validate_maybe_enum(schema, Schema, "schema")
        validate_maybe_enum(encoding, Encoding, "encoding")

        dataset = enum_or_str_lowercase(dataset, "dataset")
        schema = maybe_enum_or_str_lowercase(schema, "schema")
        encoding = maybe_enum_or_str_lowercase(encoding, "encoding")

        params: List[Tuple[str, str]] = [
            ("dataset", dataset),
            ("schema", schema),
            ("encoding", encoding),
        ]

        response: Response = self._get(
            url=self._base_url + ".list_fields",
            params=params,
            basic_auth=True,
        )
        return response.json()

    def list_encodings(self) -> List[str]:
        """
        Request all available data encodings from Databento.

        Makes a `GET /metadata.list_encodings` HTTP request.

        Returns
        -------
        List[str]

        """
        response: Response = self._get(
            url=self._base_url + ".list_encodings",
            basic_auth=True,
        )
        return response.json()

    def list_compressions(self) -> List[str]:
        """
        Request all available data compression modes from Databento.

        Makes a `GET /metadata.list_compressions` HTTP request.

        Returns
        -------
        List[str]

        """
        response: Response = self._get(
            url=self._base_url + ".list_compressions",
            basic_auth=True,
        )
        return response.json()

    def list_unit_prices(
        self,
        dataset: Union[Dataset, str],
        mode: Optional[Union[FeedMode, str]] = None,
        schema: Optional[Union[Schema, str]] = None,
    ) -> Union[float, Dict[str, Any]]:
        """
        List unit prices for each data schema in dollars per gigabyte.

        Makes a `GET /metadata.list_unit_prices` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code for the request.
        mode : FeedMode or str {'live', 'historical-streaming', 'historical'}, optional
            The data feed mode for the request.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, optional  # noqa
            The data record schema for the request.

        Returns
        -------
        float or Dict[str, Any]
            If both `mode` and `schema` are specified, the unit price is returned as a single number.
            Otherwise, return a map of feed mode to schema to unit price.

        """
        validate_maybe_enum(schema, Schema, "schema")
        validate_maybe_enum(mode, FeedMode, "mode")

        dataset = enum_or_str_lowercase(dataset, "dataset")
        mode = maybe_enum_or_str_lowercase(mode, "mode")
        schema = maybe_enum_or_str_lowercase(schema, "schema")

        params: List[Tuple[str, str]] = [
            ("dataset", dataset),
            ("mode", mode),
            ("schema", schema),
        ]

        response: Response = self._get(
            url=self._base_url + ".list_unit_prices",
            params=params,
            basic_auth=True,
        )
        return response.json()

    def get_shape(
        self,
        dataset: Union[Dataset, str],
        symbols: Optional[Union[List[str], str]] = None,
        schema: Union[Schema, str] = "trades",
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        encoding: Union[Encoding, str] = "dbz",
        stype_in: Optional[Union[SType, str]] = "native",
        limit: Optional[int] = None,
    ) -> Tuple:
        """
        Request the shape of the time series data from Databento.

        Makes a GET `/metadata.get_shape` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        encoding : Encoding or str {'dbz', 'csv', 'json'}, optional
            The data encoding.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        limit : int, optional
            The maximum number of records to include in the query.
            If ``None`` then no limit.

        Returns
        -------
        Tuple
            The shape of the data expressed as size per dimension.

        """
        validate_enum(schema, Schema, "schema")
        validate_enum(stype_in, SType, "stype_in")

        dataset = enum_or_str_lowercase(dataset, "dataset")
        symbols = maybe_symbols_list_to_string(symbols)
        schema = Schema(schema)
        start = maybe_datetime_to_string(start)
        end = maybe_datetime_to_string(end)
        encoding = Encoding(encoding)
        stype_in = SType(stype_in)

        params: List[Tuple[str, str]] = [
            ("dataset", dataset),
            ("symbols", symbols),
            ("schema", schema.value),
            ("start", start),
            ("end", end),
            ("encoding", encoding.value),
            ("stype_in", stype_in.value),
        ]
        if limit is not None:
            params.append(("limit", str(limit)))

        response: Response = self._get(
            url=self._base_url + ".get_shape",
            params=params,
            basic_auth=True,
        )

        values = response.json()
        return tuple(values)

    def get_billable_size(
        self,
        dataset: Union[Dataset, str],
        symbols: Optional[Union[List[str], str]] = None,
        schema: Union[Schema, str] = "trades",
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        stype_in: Optional[Union[SType, str]] = "native",
        limit: Optional[int] = None,
    ) -> int:
        """
        Request the billable uncompressed raw binary size for historical
        streaming or batched files from Databento.

        Makes a GET `/metadata.get_billable_size` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        limit : int, optional
            The maximum number of records to include in the query. If ``None`` then no limit.

        Returns
        -------
        int
            The size in number of bytes used for billing.

        """
        validate_enum(schema, Schema, "schema")
        validate_enum(stype_in, SType, "stype_in")

        schema = Schema(schema)
        stype_in = SType(stype_in)

        params: List[Tuple[str, str]] = super()._timeseries_params(
            dataset=dataset,
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            stype_in=stype_in,
            limit=limit,
        )

        response: Response = self._get(
            url=self._base_url + ".get_billable_size",
            params=params,
            basic_auth=True,
        )

        return response.json()

    def get_cost(
        self,
        dataset: Union[Dataset, str],
        mode: Union[FeedMode, str] = "historical-streaming",
        symbols: Optional[Union[List[str], str]] = None,
        schema: Union[Schema, str] = "trades",
        start: Optional[Union[pd.Timestamp, date, str, int]] = None,
        end: Optional[Union[pd.Timestamp, date, str, int]] = None,
        stype_in: Optional[Union[SType, str]] = "native",
        limit: Optional[int] = None,
    ) -> float:
        """
        Request the cost in US Dollars for historical streaming or batched files
        from Databento.

        Makes a `GET /metadata.get_cost` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code for the request.
        mode : FeedMode or str {'live', 'historical-streaming', 'historical'}, default 'historical-streaming'
            The data feed mode for the request.
        symbols : List[Union[str, int]] or str, optional
            The product symbols to filter for. Takes up to 2,000 symbols per request.
            If `*` or ``None`` then will be for **all** symbols.
        schema : Schema or str {'mbo', 'mbp-1', 'mbp-10', 'trades', 'tbbo', 'ohlcv-1s', 'ohlcv-1m', 'ohlcv-1h', 'ohlcv-1d', 'definition', 'statistics', 'status'}, default 'trades'  # noqa
            The data record schema for the request.
        start : pd.Timestamp or date or str or int, optional
            The start datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        end : pd.Timestamp or date or str or int, optional
            The end datetime for the request range. Assumes UTC as timezone unless otherwise specified.
            If an integer is passed, then this represents nanoseconds since UNIX epoch.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        limit : int, optional
            The maximum number of records to include in the query. If ``None`` then no limit.

        Returns
        -------
        float
            The cost in US Dollars.

        """
        validate_enum(mode, FeedMode, "mode")
        validate_enum(schema, Schema, "schema")
        validate_enum(stype_in, SType, "stype_in")

        mode = FeedMode(mode)
        schema = Schema(schema)
        stype_in = SType(stype_in)

        params: List[Tuple[str, str]] = super()._timeseries_params(
            dataset=dataset,
            symbols=symbols,
            schema=schema,
            start=start,
            end=end,
            stype_in=stype_in,
            limit=limit,
        )

        params.append(("mode", mode.value))

        response: Response = self._get(
            url=self._base_url + ".get_cost",
            params=params,
            basic_auth=True,
        )

        return response.json()



import io
import os.path
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import zstandard
from databento.common.data import DBZ_COLUMNS, DBZ_STRUCT_MAP, DERIV_SCHEMAS
from databento.common.enums import Compression, Encoding, Schema, SType
from databento.common.logging import log_debug
from databento.common.metadata import MetadataDecoder


class Bento:
    """The abstract base class for all Bento I/O classes."""

    def __init__(self):
        self._metadata: Dict[str, Any] = {}
        self._dtype: Optional[np.dtype] = None

        self._dataset: Optional[str] = None
        self._schema: Optional[Schema] = None
        self._symbols: Optional[List[str]] = None
        self._stype_in: Optional[SType] = None
        self._stype_out: Optional[SType] = None
        self._start: Optional[pd.Timestamp] = None
        self._end: Optional[pd.Timestamp] = None
        self._limit: Optional[int] = None
        self._encoding: Optional[Encoding] = None
        self._compression: Optional[Compression] = None
        self._shape: Optional[Tuple] = None

    def _check_metadata(self) -> None:
        if not self._metadata:
            self._metadata = self.source_metadata()
            if not self._metadata:
                raise RuntimeError("invalid metadata")

    def _get_index_column(self) -> str:
        return (
            "ts_event"
            if self.schema
            in (
                Schema.OHLCV_1S,
                Schema.OHLCV_1M,
                Schema.OHLCV_1H,
                Schema.OHLCV_1D,
            )
            else "ts_recv"
        )

    def source_metadata(self) -> Dict[str, Any]:
        """
        Return the metadata parsed from the data header.

        Returns
        -------
        Dict[str, Any

        """
        log_debug("Decoding metadata...")
        metadata_initial: bytes = self.reader().read(8)
        magic_bin = metadata_initial[:4]
        frame_size_bin = metadata_initial[4:]

        if not metadata_initial.startswith(b"P*M\x18"):
            return {}

        metadata_magic = int.from_bytes(bytes=magic_bin, byteorder="little")
        metadata_frame_size = int.from_bytes(bytes=frame_size_bin, byteorder="little")
        log_debug(f"magic={metadata_magic}, frame_size={metadata_frame_size}")

        metadata_raw = self.reader().read(8 + metadata_frame_size)
        return MetadataDecoder.decode_to_json(metadata_raw)

    def set_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Set metadata from a JSON object.

        Parameters
        ----------
        metadata : Dict[str, Any]
            The metadata to set.

        Warnings
        --------
        This is not intended to be called by users.

        """
        self._metadata = metadata

    def reader(self, decompress: bool = False) -> BinaryIO:
        """
        Return an I/O reader for the data.

        Parameters
        ----------
        decompress : bool
            If data should be decompressed.

        Returns
        -------
        BinaryIO

        """
        raise NotImplementedError()  # pragma: no cover

    def writer(self) -> BinaryIO:
        """
        Return an I/O writer for the data.

        Returns
        -------
        BinaryIO

        """
        raise NotImplementedError()  # pragma: no cover

    @property
    def nbytes(self) -> int:
        """
        Return the size of the data in bytes.

        Returns
        -------
        int

        """
        raise NotImplementedError()  # pragma: no cover

    @property
    def raw(self) -> bytes:
        """
        Return the raw data from the I/O stream.

        Returns
        -------
        bytes

        """
        raise NotImplementedError()  # pragma: no cover

    @property
    def dtype(self) -> np.dtype:
        """
        Return the binary struct format for the data schema.

        Returns
        -------
        np.dtype

        """
        if self._dtype is None:
            self._check_metadata()
            self._dtype = np.dtype(DBZ_STRUCT_MAP[self.schema])

        return self._dtype

    @property
    def struct_size(self) -> int:
        """
        Return the binary struct size in bytes.

        Returns
        -------
        int

        """
        return self.dtype.itemsize

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Return the metadata for the data.

        Returns
        -------
        Dict[str, Any]

        """
        return self._metadata

    @property
    def dataset(self) -> str:
        """
        Return the dataset code.

        Returns
        -------
        str

        """
        if self._dataset is None:
            self._check_metadata()
            self._dataset = self._metadata["dataset"]

        return self._dataset

    @property
    def schema(self) -> Schema:
        """
        Return the data record schema.

        Returns
        -------
        Schema

        """
        if self._schema is None:
            self._check_metadata()
            self._schema = Schema(self._metadata["schema"])

        return self._schema

    @property
    def symbols(self) -> List[str]:
        """
        Return the query symbols for the data.

        Returns
        -------
        List[str]

        """
        if self._symbols is None:
            self._check_metadata()
            self._symbols = self._metadata["symbols"]

        return self._symbols

    @property
    def stype_in(self) -> SType:
        """
        Return the query input symbology type for the data.

        Returns
        -------
        SType

        """
        if self._stype_in is None:
            self._check_metadata()
            self._stype_in = SType(self._metadata["stype_in"])

        return self._stype_in

    @property
    def stype_out(self) -> SType:
        """
        Return the query output symbology type for the data.

        Returns
        -------
        SType

        """
        if self._stype_out is None:
            self._check_metadata()
            self._stype_out = SType(self._metadata["stype_out"])

        return self._stype_out

    @property
    def start(self) -> pd.Timestamp:
        """
        Return the query start for the data.

        Returns
        -------
        pd.Timestamp

        Notes
        -----
        The data timestamps will not occur prior to `start`.

        """
        if self._start is None:
            self._check_metadata()
            self._start = pd.Timestamp(self._metadata["start"], tz="UTC")

        return self._start

    @property
    def end(self) -> pd.Timestamp:
        """
        Return the query end for the data.

        Returns
        -------
        pd.Timestamp

        Notes
        -----
        The data timestamps will not occur after `end`.

        """
        if self._end is None:
            self._check_metadata()
            self._end = pd.Timestamp(self._metadata["end"], tz="UTC")

        return self._end

    @property
    def limit(self) -> Optional[int]:
        """
        Return the query limit for the data.

        Returns
        -------
        int or None

        """
        if self._limit is None:
            self._check_metadata()
            self._limit = self._metadata["limit"]

        return self._limit

    @property
    def compression(self) -> Compression:
        """
        Return the data compression format (if any).

        Returns
        -------
        Compression

        """
        if self._compression is None:
            self._check_metadata()
            self._compression = Compression(self._metadata["compression"])

        return self._compression

    @property
    def shape(self) -> Tuple:
        """
        Return the shape of the data.

        Returns
        -------
        Tuple
            The data shape.

        """
        if self._shape is None:
            self._check_metadata()
            self._shape = (
                self._metadata["record_count"],
                len(DBZ_STRUCT_MAP[self.schema]),
            )

        return self._shape

    @property
    def mappings(self) -> List[Dict[str, List[Dict[str, str]]]]:
        """
        Return the symbology mappings for the data.

        Returns
        -------
        List[Dict[str, List[Dict[str, str]]]]

        """
        self._check_metadata()

        return self._metadata["mappings"]

    @property
    def symbology(self) -> Dict[str, Any]:
        """
        Return the symbology resolution information for the query.

        Returns
        -------
        Dict[str, Any]

        """
        self._check_metadata()

        status = 0
        if self._metadata["partial"]:
            status = 1
            message = "Partially resolved"
        elif self._metadata["not_found"]:
            status = 2
            message = "Not found"
        else:
            message = "OK"

        response: Dict[str, Any] = {
            "result": self.mappings,
            "symbols": self.symbols,
            "stype_in": self.stype_in.value,
            "stype_out": self.stype_out.value,
            "start_date": str(self.start.date()),
            "end_date": str(self.end.date()),
            "partial": self._metadata["partial"],
            "not_found": self._metadata["not_found"],
            "message": message,
            "status": status,
        }

        return response

    def to_ndarray(self) -> np.ndarray:
        """
        Return the data as a numpy `ndarray`.

        Returns
        -------
        np.ndarray

        """
        data: bytes = self.reader(decompress=True).read()
        return np.frombuffer(data, dtype=DBZ_STRUCT_MAP[self.schema])

    def to_df(self, pretty_ts: bool = False, pretty_px: bool = False) -> pd.DataFrame:
        """
        Return the data as a `pd.DataFrame`.

        Parameters
        ----------
        pretty_ts : bool, default False
            If all timestamp columns should be converted from UNIX nanosecond
            `int` to `pd.Timestamp` tz-aware (UTC).
        pretty_px : bool, default False
            If all price columns should be converted from `int` to `float` at
            the correct scale (using the fixed precision scalar 1e-9).

        Returns
        -------
        pd.DataFrame

        """
        df = pd.DataFrame(self.to_ndarray())
        df.set_index(self._get_index_column(), inplace=True)

        # Cleanup dataframe
        if self.schema == Schema.MBO:
            df.drop("channel_id", axis=1, inplace=True)
            df = df.reindex(columns=DBZ_COLUMNS[self.schema])
            df["flags"] = df["flags"] & 0xFF  # Apply bitmask
            df["side"] = df["side"].str.decode("utf-8")
            df["action"] = df["action"].str.decode("utf-8")
        elif self.schema in DERIV_SCHEMAS:
            df.drop(["nwords", "type", "depth"], axis=1, inplace=True)
            df = df.reindex(columns=DBZ_COLUMNS[self.schema])
            df["flags"] = df["flags"] & 0xFF  # Apply bitmask
            df["side"] = df["side"].str.decode("utf-8")
            df["action"] = df["action"].str.decode("utf-8")
        else:
            df.drop(["nwords", "type"], axis=1, inplace=True)

        if pretty_ts:
            df.index = pd.to_datetime(df.index, utc=True)
            for column in df.columns:
                if column.startswith("ts_") and "delta" not in column:
                    df[column] = pd.to_datetime(df[column], utc=True)

        if pretty_px:
            for column in list(df.columns):
                if (
                    column in ("price", "open", "high", "low", "close")
                    or column.startswith("bid_px")  # MBP
                    or column.startswith("ask_px")  # MBP
                ):
                    df[column] = df[column] * 1e-9

        return df

    def replay(self, callback: Callable[[Any], None]) -> None:
        """
        Replay data by passing records sequentially to the given callback.

        Parameters
        ----------
        callback : callable
            The callback to the data handler.

        """
        dtype = DBZ_STRUCT_MAP[self.schema]
        reader: BinaryIO = self.reader(decompress=True)
        while True:
            raw: bytes = reader.read(self.struct_size)
            record = np.frombuffer(raw, dtype=dtype)
            if record.size == 0:
                break
            callback(record[0])

    @staticmethod
    def from_file(path: str) -> "FileBento":
        """
        Load the data from a DBZ file at the given path.

        Parameters
        ----------
        path : str
            The path to read from.

        Returns
        -------
        FileBento

        Raises
        ------
        FileNotFoundError
            If no file is found at the given path.
        RuntimeError
            If an empty file exists at the given path.

        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"no file found at `path` '{path}'")
        if os.stat(path).st_size == 0:
            raise RuntimeError(f"the file at `path` '{path}' was empty")

        bento = FileBento(path=path)

        metadata = bento.source_metadata()
        bento.set_metadata(metadata)

        return bento

    def to_file(self, path: str) -> "FileBento":
        """
        Write the data to a DBZ file at the given path.

        Parameters
        ----------
        path : str
            The file path to write to.

        Returns
        -------
        FileBento

        """
        with open(path, mode="wb") as f:
            f.write(self.reader().read())

        bento = FileBento(path=path)
        bento.set_metadata(self._metadata)

        return bento

    def to_csv(self, path: str) -> None:
        """
        Write the data to a file in CSV format.

        Parameters
        ----------
        path : str
            The file path to write to.

        Notes
        -----
        Requires all the data to be brought up into memory to then be written.

        """
        self.to_df().to_csv(path)

    def to_json(self, path: str) -> None:
        """
        Write the data to a file in JSON format.

        Parameters
        ----------
        path : str
            The file path to write to.

        Notes
        -----
        Requires all the data to be brought up into memory to then be written.

        """
        self.to_df().to_json(path, orient="records", lines=True)

    def request_symbology(self, client) -> Dict[str, Any]:
        """
        Request symbology resolution based on the metadata properties.

        Makes a `GET /symbology.resolve` HTTP request.

        Current symbology mappings from the metadata are also available by
        calling the `.symbology` or `.mappings` properties.

        Parameters
        ----------
        client : Historical
            The historical client to use for the request.

        Returns
        -------
        Dict[str, Any]
            A result including a map of input symbol to output symbol across a
            date range.

        """
        return client.symbology.resolve(
            dataset=self.dataset,
            symbols=self.symbols,
            stype_in=self.stype_in,
            stype_out=self.stype_out,
            start_date=self.start.date(),
            end_date=self.end.date(),
        )

    def request_full_definitions(
        self,
        client,
        path: Optional[str] = None,
    ) -> "Bento":
        """
        Request full instrument definitions based on the metadata properties.

        Makes a `GET /timeseries.stream` HTTP request.

        Parameters
        ----------
        client : Historical
            The historical client to use for the request (contains the API key).
        path : str, optional
            The path to stream the data to on disk (will then return a `FileBento`).

        Returns
        -------
        Bento

        Warnings
        --------
        Calling this method will incur a cost.

        """
        return client.timeseries.stream(
            dataset=self.dataset,
            symbols=self.symbols,
            schema=Schema.DEFINITION,
            start=self.start,
            end=self.end,
            stype_in=self.stype_in,
            stype_out=self.stype_out,
            path=path,
        )


class MemoryBento(Bento):
    """
    Provides data streaming I/O operations backed by an in-memory buffer.

    Parameters
    ----------
    initial_bytes : bytes, optional
        The initial data for the memory buffer.
    """

    def __init__(self, initial_bytes: Optional[bytes] = None):
        super().__init__()

        self._buffer = io.BytesIO(initial_bytes=initial_bytes or b"")

    @property
    def nbytes(self) -> int:
        return self._buffer.getbuffer().nbytes

    @property
    def raw(self) -> bytes:
        return self._buffer.getvalue()

    def reader(self, decompress: bool = False) -> BinaryIO:
        self._buffer.seek(0)  # Ensure reader at start of stream
        if decompress:
            return zstandard.ZstdDecompressor().stream_reader(self._buffer.getbuffer())
        else:
            return self._buffer

    def writer(self) -> BinaryIO:
        return self._buffer


class FileBento(Bento):
    """
    Provides data streaming I/O operations backed by a file on disk.

    Parameters
    ----------
    path : str
        The path to the data file.
    """

    def __init__(self, path: str):
        super().__init__()

        self._path = path

    @property
    def path(self) -> str:
        """
        Return the path to the backing data file.

        Returns
        -------
        str

        """
        return self._path

    @property
    def nbytes(self) -> int:
        return os.path.getsize(self._path)

    @property
    def raw(self) -> bytes:
        return self.reader().read()

    def reader(self, decompress: bool = False) -> BinaryIO:
        f = open(self._path, mode="rb")
        if decompress:
            return zstandard.ZstdDecompressor().stream_reader(f)
        else:
            return f

    def writer(self) -> BinaryIO:
        return open(self._path, mode="wb")



from enum import Enum, EnumMeta
from typing import List, Optional, Union


def validate_enum(
    value: Union[Enum, str],
    enum: EnumMeta,
    param: str,
    lower: bool = True,
) -> Union[Enum, str]:
    """
    Validate whether the given value is either the correct Enum type, or a valid
    value of that enum.

    Parameters
    ----------
    value : Enum or str
        The value to validate.
    enum : EnumMeta
        The valid enum type.
    param : str
        The name of the parameter being validated (for any error message).
    lower : str, default True
        If the str value should be lower-cased.

    Returns
    -------
    Union[Enum, str]
        The valid enum or lower-cased string.

    Raises
    ------
    ValueError
        If value is not valid for the given enum.
    TypeError
        If value is an Enum and not of type enum_type.

    """
    if isinstance(value, str):
        if lower:
            value = value.lower()
        valid: List[str] = [x.value for x in enum]  # type: ignore
        if value not in valid:
            raise ValueError(
                f"The `{param}` was not a valid value of {enum}, was '{value}'. "
                f"Use any of {valid}.",
            )
    elif not isinstance(value, enum):
        raise TypeError(
            f"The `{param}` value was not of type {enum}, was {type(value)}.",
        )

    return value


def validate_maybe_enum(
    value: Optional[Union[Enum, str]],
    enum: EnumMeta,
    param: str,
    lower: bool = True,
) -> Optional[Union[Enum, str]]:
    """
    Validate whether the given value is either the correct Enum type, a valid
    value of that enum, or None.

    Parameters
    ----------
    value : Enum or str, optional
        The value to validate.
    enum : EnumMeta
        The valid enum type.
    param : str
        The name of the parameter being validated (for any error message).
    lower : str, default True
        If the str value should be lower-cased.

    Returns
    -------
    Union[Enum, str] or None
        The valid enum or lower-cased string.

    Raises
    ------
    ValueError
        If value is not valid for the given enum.
    TypeError
        If value is an Enum and not of type enum_type.

    """
    if value is None:
        return None
    return validate_enum(value, enum, param, lower)



from datetime import date
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from databento.common.enums import SType
from databento.common.parsing import enum_or_str_lowercase, maybe_symbols_list_to_string
from databento.historical.api import API_VERSION
from databento.historical.http import BentoHttpAPI
from requests import Response


class SymbologyHttpAPI(BentoHttpAPI):
    """
    Provides request methods for the symbology HTTP API endpoints.
    """

    def __init__(self, key, gateway):
        super().__init__(key=key, gateway=gateway)
        self._base_url = gateway + f"/v{API_VERSION}/symbology"

    def resolve(
        self,
        dataset: str,
        symbols: Union[List[str], str],
        stype_in: Union[SType, str],
        stype_out: Union[SType, str],
        start_date: Union[date, str],
        end_date: Union[date, str],
        default_value: Optional[str] = "",
    ) -> Dict[str, Any]:
        """
        Request symbology mappings resolution from Databento.

        Makes a `GET /symbology.resolve` HTTP request.

        Parameters
        ----------
        dataset : Dataset or str
            The dataset code (string identifier) for the request.
        symbols : List[Union[str, int]] or str, optional
            The symbols to resolve. Takes up to 2,000 symbols per request.
            If `*` or ``None`` then will be for **all** symbols.
        stype_in : SType or str, default 'native'
            The input symbology type to resolve from.
        stype_out : SType or str, default 'product_id'
            The output symbology type to resolve to.
        start_date : date or str
            The start date (UTC) of the request time range (inclusive).
        end_date : date or str
            The end date (UTC) of the request time range (exclusive).
        default_value : str, default '' (empty string)
            The default value to return if a symbol cannot be resolved.

        Returns
        -------
        Dict[str, Any]
            A result including a map of input symbol to output symbol across a
            date range.

        """
        dataset = enum_or_str_lowercase(dataset, "dataset")
        symbols = maybe_symbols_list_to_string(symbols)
        stype_in = enum_or_str_lowercase(stype_in, "stype_in")
        stype_out = enum_or_str_lowercase(stype_out, "stype_out")
        start_date = str(pd.to_datetime(start_date).date())
        end_date = str(pd.to_datetime(end_date).date())

        params: List[Tuple[str, str]] = [
            ("dataset", dataset),
            ("symbols", symbols),
            ("stype_in", stype_in),
            ("stype_out", stype_out),
            ("start_date", start_date),
            ("end_date", end_date),
            ("default_value", default_value),
        ]

        response: Response = self._get(
            url=self._base_url + ".resolve",
            params=params,
            basic_auth=True,
        )

        return response.json()



class BentoError(Exception):
    """
    Represents a Databento specific error.
    """

    pass


class BentoHttpError(BentoError):
    """
    Represents a Databento specific HTTP error.
    """

    def __init__(
        self,
        http_status=None,
        http_body=None,
        json_body=None,
        message=None,
        headers=None,
    ):
        super(BentoHttpError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "<Could not decode body as utf-8. "
                    "Please report to support@databento.com>"
                )

        self.http_status = http_status
        self.http_body = http_body
        self.json_body = json_body
        self.message = message
        self.headers = headers or {}
        self.request_id = self.headers.get("request-id", None)

    def __str__(self):
        msg = self.message or "<empty message>"
        msg = f"{self.http_status} {msg}"
        if self.request_id is not None:
            return f"Request {self.request_id}: {msg}"
        else:
            return msg

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"request_id={self.request_id}, "
            f"http_status={self.http_status}, "
            f"message={self.message})"
        )


class BentoServerError(BentoHttpError):
    """
    Represents a Databento specific server side 500 series HTTP error.
    """

    def __init__(
        self,
        http_status=None,
        http_body=None,
        json_body=None,
        message=None,
        headers=None,
    ):
        super().__init__(
            http_body=http_body,
            http_status=http_status,
            json_body=json_body,
            message=message,
            headers=headers,
        )


class BentoClientError(BentoHttpError):
    """
    Represents a Databento specific client side 400 series HTTP error.
    """

    def __init__(
        self,
        http_status=None,
        http_body=None,
        json_body=None,
        message=None,
        headers=None,
    ):
        super().__init__(
            http_body=http_body,
            http_status=http_status,
            json_body=json_body,
            message=message,
            headers=headers,
        )



import databento as db
from databento import Bento
from databento.common.enums import Schema


if __name__ == "__main__":
    key = ""  # <-- Change to valid API key prior to running
    client = db.Historical(key=key)

    for schema in Schema:
        if schema in (Schema.DEFINITION, Schema.STATISTICS, Schema.STATUS):
            continue

        print(schema.value)

        path = f"test_data.{schema.value}.dbz"

        # Execute request through client
        data: Bento = client.timeseries.stream(
            dataset="GLBX.MDP3",
            symbols=["ESH1"],
            schema=schema,
            start="2020-12-28T13:00",
            end="2020-12-29",
            limit=2,  # <-- limiting response to 2 records only (for test cases)
            path=path,
        )  # -> FileBento

        written = open(path, mode="rb").read()
        print(written)



#!/usr/bin/env python3
import os
from typing import Any, Dict

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))
os.chdir(here)


with open("README.md") as f:
    readme = f.read()

# Option 3 for single-sourcing version information, via
# https://packaging.python.org/en/latest/guides/single-sourcing-package-version/
with open("databento/version.py") as f:
    version_globals: Dict[str, Any] = {}
    exec(f.read(), version_globals)
    version = version_globals["__version__"]

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines(keepends=False)

with open("requirements_dev.txt", "r") as f:
    tests_require = f.read().splitlines(keepends=False)


setup(
    name="databento",
    version=version,
    description="Official Python client library for Databento",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Databento",
    author_email="support@databento.com",
    url="https://github.com/databento/databento-python",
    license="Apache License 2.0",
    keywords="databento financial data API",
    packages=find_packages(exclude=["tests", "examples"]),
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    python_requires=">=3.7.*",
    project_urls={
        "Bug Tracker": "https://github.com/databento/databento-python/issues",
        "Documentation": "https://docs.databento.com/",
        "Source Code": "https://github.com/databento/databento-python",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
    ],
)
